from django.urls    import path

from .views import (UserListView, UserView, ProjectListView, ProjectView, 
                    SystemListView, SystemView, ServiceView, JobListView, JobView)

app_name = "openacct"

urlpatterns = [
    path('user_list/', UserListView.as_view(), name='user_list'),
    path('user/<int:byid>/', UserView.as_view(), name='user_byid'),
    path('user/<byname>/', UserView.as_view(), name='user_byname'),
    path('project_list/', ProjectListView.as_view(), name='project_list'),
    path('project/<int:byid>/', ProjectView.as_view(), name='project_byid'),
    path('project/<byname>/', ProjectView.as_view(), name='project_byname'),
    path('system_list/', SystemListView.as_view(), name='system_list'),
    path('system/<int:byid>/', SystemView.as_view(), name='system_byid'),
    path('system/<byname>/', SystemView.as_view(), name='system_byname'),
    path('service/<int:byid>/', ServiceView.as_view(), name='service_byid'),
    path('service/<byname>/', ServiceView.as_view(), name='service_byname'),
    path('job_list/', JobListView.as_view(), name='job_list'),
    path('job/<int:byid>/', JobView.as_view(), name='job_byid'),
    path('job/<byname>/', JobView.as_view(), name='job_byname'),
]



# Generated by Django 2.2.7 on 2020-10-31 02:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openacct', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='default_project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Project', blank=True, null=True, related_name='+'),
        ),
    ]




import re

from django.db import models


class User(models.Model):
    created     = models.DateTimeField(auto_now_add=True)
    name        = models.CharField(max_length=32, unique=True)
    realname    = models.CharField(max_length=128, blank=True)
    active      = models.BooleanField(blank=True, default=True)
    projects    = models.ManyToManyField('Project', blank=True)
    default_project = models.ForeignKey('Project', on_delete=models.CASCADE, 
        blank=True, null=True, related_name='+'
    )

    def __str__(self):
        return '<User: {}>'.format(self.name)


class Project(models.Model):
    created     = models.DateTimeField(auto_now_add=True)
    name        = models.CharField(max_length=32, unique=True)
    ldap_group  = models.CharField(max_length=32, blank=True)
    active      = models.BooleanField(blank=True, default=True)
    pi          = models.ForeignKey(User, on_delete=models.CASCADE)
    description = models.CharField(max_length=1024, blank=True, default="")
        
    def __str__(self):
        return '{} - {}'.format(self.name, self.description)

    def get_index_value(self):
        m = re.search('\d+$', self.name)
        return int(m.group(0), 10) if m else 0

    @classmethod
    def next_index(cls, prefix):
        return 1 + max([
            p.get_index_value() for p in cls.objects.filter(name__icontains=prefix)
        ] + [0])


class UserProjectEvent(models.Model):
    EVENT_TYPES = (
        ('ADDED', 'ADDED'),
        ('REMOVED', 'REMOVED'),
    )
    created     = models.DateTimeField(auto_now_add=True)
    user        = models.ForeignKey(User, on_delete=models.CASCADE)
    project     = models.ForeignKey(Project, on_delete=models.CASCADE)
    event_type  = models.CharField(max_length=16, choices=EVENT_TYPES)


class Account(models.Model):
    created     = models.DateTimeField(auto_now_add=True)
    name        = models.CharField(max_length=32)
    active      = models.BooleanField(blank=True, default=True)
    expires     = models.DateTimeField(blank=True)
    project     = models.ForeignKey(Project, on_delete=models.CASCADE)
    services    = models.ManyToManyField('Service', blank=True)

    def __str__(self):
        return '<Account: {}>'.format(self.name)

    def get_index_value(self):
        m = re.search('\d+$', self.name)
        return int(m.group(0), 10) if m else 0

    @classmethod
    def next_index(cls, prefix):
        return 1 + max([
            a.get_index_value() for a in cls.objects.filter(name__icontains=prefix)
        ] + [0])


class System(models.Model):
    created     = models.DateTimeField(auto_now_add=True)
    name        = models.CharField(max_length=32, unique=True)
    active      = models.BooleanField(blank=True, default=True)
    description = models.CharField(max_length=1024, blank=True, default="")

    def __str__(self):
        return '<System: {}>'.format(self.name)


class Service(models.Model):
    created     = models.DateTimeField(auto_now_add=True)
    name        = models.CharField(max_length=32)
    units       = models.CharField(max_length=32)
    active      = models.BooleanField(blank=True, default=True)
    system      = models.ForeignKey(System, on_delete=models.CASCADE)
    charge_rate = models.FloatField(blank=True, default=0)
    description = models.CharField(max_length=1024, blank=True, default="")

    def __str__(self):
        return '<Service: {}>'.format(self.name)


class Transaction(models.Model):
    TX_TYPES = (
        ('AUDIT', 'AUDIT'),
        ('CREDIT', 'CREDIT'),
        ('DEBIT', 'DEBIT'),
        ('GRANT', 'GRANT'),
        ('REVOKE', 'REVOKE'),
    )
    created     = models.DateTimeField(auto_now_add=True)
    active      = models.BooleanField(blank=True, default=True)
    service     = models.ForeignKey(Service, on_delete=models.CASCADE)
    account     = models.ForeignKey(Account, on_delete=models.CASCADE)
    creator     = models.ForeignKey(User, on_delete=models.CASCADE)
    amt_used    = models.FloatField()
    amt_charged = models.FloatField(blank=True, default=0.0)
    tx_type     = models.CharField(max_length=16, choices=TX_TYPES)


class Job(models.Model):
    queued      = models.DateTimeField()
    started     = models.DateTimeField(blank=True, null=True)
    completed   = models.DateTimeField(blank=True, null=True)
    jobid       = models.CharField(max_length=32, unique=True)
    name        = models.CharField(max_length=64, blank=True, default='')
    submit_host = models.CharField(max_length=64, blank=True, default='')
    host_list   = models.CharField(max_length=1024, blank=True, default='')
    qos         = models.CharField(max_length=64, blank=True, default='')
    job_script  = models.TextField(blank=True, default='')
    transactions   = models.ManyToManyField(Transaction, blank=True)
    wall_requested = models.IntegerField()
    wall_duration  = models.IntegerField(blank=True)

    def __str__(self):
        return '<Job: {} - {}>'.format(self.jobid, self.name)


class StorageCommitment(models.Model):
    DIR_TYPES = (
        ('HOME', 'Home Directory'),
        ('PROJECT', 'Project Space'),
        ('SCRATCH', 'Scratch Space'),
        ('TEMP', 'Temporary Space'),
    )
    dir_type        = models.CharField(max_length=16, choices=DIR_TYPES)
    project         = models.ForeignKey(Project, on_delete=models.CASCADE)
    filesystem      = models.CharField(max_length=64)
    path            = models.CharField(max_length=256)
    commitment      = models.BigIntegerField(default=0, blank=True, null=True)
    registered      = models.DateTimeField(auto_now_add=True)
    created         = models.DateTimeField(blank=True, null=True)
    end_date        = models.DateTimeField(blank=True, null=True)
    reclaimed       = models.DateTimeField(blank=True, null=True)
    uid             = models.IntegerField(blank=True, default=0)
    gid             = models.IntegerField(blank=True, default=0)
    pid             = models.IntegerField(blank=True, default=0)
    permissions     = models.CharField(max_length=8, default='0700')
    is_purged       = models.BooleanField(blank=True, default=True)
    transactions    = models.ManyToManyField(Transaction, blank=True)



__version__     = "0.0.5"

__title__       = "django-openacct"
__description__ = "A pluggable Django app for managing research computing center project accounting."
__uri__         = "https://github.com/Elemnir/django-openacct"

__author__      = "Adam Howard"
__email__       = "ahoward@utk.edu"

__license__     = "BSD 3-clause"
__copyright__   = "Copyright (c) 2020 Adam Howard"





from django.contrib import admin

from .models    import (User, Project, UserProjectEvent, Account, System, Service,
                        Transaction, Job, StorageCommitment)

from .shortcuts import add_user_to_project, create_account


class UserAdmin(admin.ModelAdmin):
    pass
admin.site.register(User, UserAdmin)


class ProjectAdmin(admin.ModelAdmin):
    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            add_user_to_project(obj.pi, obj)
            create_account(obj)

admin.site.register(Project, ProjectAdmin)


class UserProjectEventAdmin(admin.ModelAdmin):
    pass
admin.site.register(UserProjectEvent, UserProjectEventAdmin)


class AccountAdmin(admin.ModelAdmin):
    pass
admin.site.register(Account, AccountAdmin)


class SystemAdmin(admin.ModelAdmin):
    pass
admin.site.register(System, SystemAdmin)


class ServiceAdmin(admin.ModelAdmin):
    pass
admin.site.register(Service, ServiceAdmin)


class TransactionAdmin(admin.ModelAdmin):
    pass
admin.site.register(Transaction, TransactionAdmin)


class JobAdmin(admin.ModelAdmin):
    pass
admin.site.register(Job, JobAdmin)


class StorageCommitmentAdmin(admin.ModelAdmin):
    pass
admin.site.register(StorageCommitment, StorageCommitmentAdmin)




from django.forms   import ModelForm

from .models    import (User, Project, UserProjectEvent, Account, System, Service, 
                        Transaction, Job, StorageCommitment)
from .shortcuts import add_user_to_project, create_account

class ProjectForm(ModelForm):
    class Meta:
        model = Project
        fields = ['pi', 'name', 'description', 'ldap_group']

    def save(self, commit=True):
        proj = super().save(commit)
        if commit:
            add_user_to_project(proj.pi, proj)
            create_account(proj)
        return proj



from datetime   import datetime

from django.core                import serializers
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.http                import HttpResponseBadRequest, HttpResponseForbidden, JsonResponse
from django.shortcuts           import render, get_object_or_404
from django.views.generic       import View 

from .models    import User, Project, Account, System, Service, Transaction, Job


#######################################################################
#
#   Informational Views
#
#######################################################################

class UserListView(LoginRequiredMixin, UserPassesTestMixin, View):
    """Returns a list of users, supports GET parameter filters"""
    def test_func(self):
        return self.request.user.is_staff

    def get(self, request):
        filters = {'active': True} if 'active' in self.request.GET else {}
        for field in ['name', 'realname']:
            if self.request.GET.get(field, False):
                filters[field + '__icontains'] = self.request.GET[field]
        
        payload = {'users':[]}
        for user in User.objects.filter(**filters).order_by('name'):
            payload['users'].append({
                'id': user.pk,
                'name': user.name, 
                'realname': user.realname,
                'created':  user.created,
            })
        return JsonResponse(payload)


class UserView(LoginRequiredMixin, View):
    """Returns a specific user"""
    def get(self, request, byid=None, byname=None):
        if (byid == None and byname == None) or (byid != None and byname != None):
            return HttpResponseBadRequest()
        
        user = (get_object_or_404(User, pk=byid) if byid 
            else get_object_or_404(User, name=byname))
        
        if not self.request.user.is_staff and self.request.username != user.name:
            return HttpResponseForbidden()
        
        return JsonResponse({'user': {
            'id': user.pk,
            'name': user.name, 
            'realname': user.realname,
            'created':  user.created,
            'projects': [ {
                'id': p.pk, 'name': p.name, 
                'ldap_group': p.ldap_group, 
                'description': p.description
            } for p in user.projects.filter(active=True)]
        }})


class ProjectListView(LoginRequiredMixin, UserPassesTestMixin, View):
    """Returns a list of projects. Supports GET parameter filters"""
    def test_func(self):
        return self.request.user.is_staff

    def get(self, request):
        filters = {'active': True} if 'active' in self.request.GET else {}
        for field in ['name', 'description']:
            if self.request.GET.get(field, False):
                filters[field + '__icontains'] = self.request.GET[field]
        
        payload = {'projects': []}
        for project in Project.objects.filter(**filters).order_by('name'):
            payload['projects'].append({
                'id': project.pk,
                'name': project.name,
                'pi': project.pi.name,
                'ldap_group': project.ldap_group,
                'description': project.description,
                'created': project.created,
            })
        return JsonResponse(payload)


class ProjectView(LoginRequiredMixin, View):
    """Returns a specific project"""
    def get(self, request, byid=None, byname=None):
        if (byid == None and byname == None) or (byid != None and byname != None):
            return HttpResponseBadRequest()
        
        project = (get_object_or_404(Project, pk=byid) if byid 
            else get_object_or_404(Project, name=byname))
        
        if (not self.request.user.is_staff and self.request.user.username 
                not in [u.name for u in project.user_set]):
            return HttpResponseForbidden()

        return JsonResponse({'project': {
            'id': project.pk,
            'name': project.name,
            'pi': project.pi.name,
            'ldap_group': project.ldap_group,
            'description': project.description,
            'created': project.created,
            'users': [ {
                'id': u.pk, 'name': u.name,
                'realname': u.realname,
                'created': u.created
            } for u in project.user_set.all() if u.active ],
            'accounts': [ {
                'id': a.pk, 'name': a.name, 
                'created': a.created,
                'expires': a.expires,
            } for a in Account.objects.filter(project=project, active=True)],
        }})
        

class SystemListView(LoginRequiredMixin, UserPassesTestMixin, View):
    """Returns a list of systems. Supports GET parameter filters"""
    def test_func(self):
        return self.request.user.is_staff

    def get(self, request):
        filters = {'active': True} if 'active' in self.request.GET else {}
        for field in ['name', 'description']:
            if self.request.GET.get(field, False):
                filters[field + '__icontains'] = self.request.GET[field]
        
        payload = {'systems': []}
        for system in System.objects.filter(**filters).order_by('name'):
            payload['systems'].append({
                'id': system.pk,
                'name': system.name,
                'description': system.description,
                'created': system.created,
            })
        return JsonResponse(payload)


class SystemView(LoginRequiredMixin, View):
    """Returns a specific system and services"""
    def get(self, request, byid=None, byname=None):
        if (byid == None and byname == None) or (byid != None and byname != None):
            return HttpResponseBadRequest()
        
        system = (get_object_or_404(System, pk=byid) if byid 
            else get_object_or_404(System, name=byname))

        return JsonResponse({'system': {
            'id': system.pk,
            'name': system.name,
            'description': system.description,
            'created': system.created,
            'services': [ {
                'id': s.pk, 'name': s.name, 'units': s.units,
                'charge_rate': s.charge_rate,
                'description': s.description,
            } for s in Service.objects.filter(system=system, active=True)],
        }})


class ServiceView(LoginRequiredMixin, UserPassesTestMixin, View):
    """Returns a specific service and accounts allowed to use it"""
    def test_func(self):
        return self.request.user.is_staff

    def get(self, request, byid=None, byname=None):
        if (byid == None and byname == None) or (byid != None and byname != None):
            return HttpResponseBadRequest()
        
        service = (get_object_or_404(Service, pk=byid) if byid 
            else get_object_or_404(Service, name=byname))

        return JsonResponse({'service': {
            'id': service.pk,
            'name': service.name,
            'description': service.description,
            'system': service.system.name,
            'units': service.units,
            'charge_rate': service.charge_rate,
            'created': service.created,
            'accounts': [ {
                'id': a.pk, 'name': a.name,
                'project': a.project.name,
                'created': a.created,
                'expires': a.expires,
            } for a in Service.account_set.filter(active=True)],
        }})


class JobListView(LoginRequiredMixin, UserPassesTestMixin, View):
    """Returns a list of jobs. Supports GET parameter filters"""
    def test_func(self):
        return self.request.user.is_staff

    def get(self, request):
        filters = {}
        for field in ['name','jobid','submit_host','host_list','qos','job_script']:
            if self.request.GET.get(field, False):
                filters[field + '__icontains'] = self.request.GET[field]
        
        for field in ['queued_after', 'started_after', 'completed_after']:
            if self.request.GET.get(field, False):
                filters[field.replace('_after', '__gte')] = datetime.fromisoformat(
                    self.request.GET[field])
        
        for field in ['queued_before', 'started_before', 'completed_before']:
            if self.request.GET.get(field, False):
                filters[field.replace('_before', '__lte')] = datetime.fromisoformat(
                    self.request.GET[field])
        
        jobs = Job.objects.filter(**filters).order_by('jobid')
        
        # These filters are EXPENSIVE
        txfilters = {}
        for field in ['account', 'service', 'creator']:
            if self.request.GET.get(field, False):
                txfilters[field+'__name__icontains'] = self.request.GET[field]

        if txfilters:
            txfilters['active'] = True
            txids = Transaction.objects.filter(**txfilters).values_list('id', flat=True)
            jobs.filter(transactions__in=txids)
                
        payload = {'jobs': []}
        for job in jobs:
            payload['jobs'].append({
                'id': job.pk, 'jobid': job.jobid, 'name': job.name, 'qos': job.qos,
                'submit_host': job.submit_host, 'host_list': job.host_list,
                'queued': job.queued, 
                'started': job.started, 
                'completed': job.completed, 
                'wall_requested': job.wall_requested,
                'wall_duration': job.wall_duration, 
                'transactions': ( [ {
                    'id': t.pk, 'tx_type': t.tx_type, 'created': t.created, 
                    'amt_used': t.amt_used, 'amt_charged': t.amt_charged,
                    'service': t.service.name,
                    'account': t.account.name,
                    'creator': t.creator.name,
                } for t in job.transactions.filter(active=True).order_by('-created')] 
                if self.request.GET.get('show_txs', False) else [] )
            })
        return JsonResponse(payload)


class JobView(LoginRequiredMixin, View):
    """Returns a specific job"""
    def get(self, request, byid=None, byname=None):
        if (byid == None and byname == None) or (byid != None and byname != None):
            return HttpResponseBadRequest()
        
        job = (get_object_or_404(Job, pk=byid) if byid 
            else get_object_or_404(Job, name=byname))
        
        if (not self.request.user.is_staff and self.request.user.username 
                not in [t.creator.name for t in job.transactions]):
            return HttpResponseForbidden()

        return JsonResponse({'job' : {
            'id': job.pk, 'jobid': job.jobid, 'name': job.name, 'qos': job.qos,
            'submit_host': job.submit_host, 'host_list': job.host_list,
            'queued': job.queued, 'started': job.started, 'completed': job.completed, 
            'wall_requested': job.wall_requested, 'wall_duration': job.wall_duration, 
            'job_script': job.job_script,
            'transactions': [ {
                'id': t.pk, 'tx_type': t.tx_type, 'created': t.created, 
                'amt_used': t.amt_used, 'amt_charged': t.amt_charged,
                'service': t.service.name,
                'account': t.account.name,
                'creator': t.creator.name,
            } for t in job.transactions.filter(active=True).order_by('-created')]
        }})
        

#######################################################################
#
#   Create/Modify Views
#
#######################################################################

class JobChangeView(LoginRequiredMixin, UserPassesTestMixin, View):
    def test_func(self):
        return self.request.user.is_staff

    def put(self, request):
        pass



# Generated by Django 2.2.7 on 2020-08-23 18:15

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=32)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('expires', models.DateTimeField(blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=32, unique=True)),
                ('ldap_group', models.CharField(blank=True, max_length=32)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('description', models.CharField(blank=True, default='', max_length=1024)),
            ],
        ),
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=32)),
                ('units', models.CharField(max_length=32)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('charge_rate', models.FloatField(blank=True, default=0)),
                ('description', models.CharField(blank=True, default='', max_length=1024)),
            ],
        ),
        migrations.CreateModel(
            name='System',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=32, unique=True)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('description', models.CharField(blank=True, default='', max_length=1024)),
            ],
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('name', models.CharField(max_length=32, unique=True)),
                ('realname', models.CharField(blank=True, max_length=128)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('projects', models.ManyToManyField(blank=True, to='openacct.Project')),
            ],
        ),
        migrations.CreateModel(
            name='UserProjectEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('event_type', models.CharField(choices=[('ADDED', 'ADDED'), ('REMOVED', 'REMOVED')], max_length=16)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.User')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Project')),
            ],
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('active', models.BooleanField(blank=True, default=True)),
                ('amt_used', models.FloatField()),
                ('amt_charged', models.FloatField(blank=True, default=0.0)),
                ('tx_type', models.CharField(choices=[('AUDIT', 'AUDIT'), ('CREDIT', 'CREDIT'), ('DEBIT', 'DEBIT'), ('GRANT', 'GRANT'), ('REVOKE', 'REVOKE')], max_length=16)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Account')),
                ('creator', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.User')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Service')),
            ],
        ),
        migrations.CreateModel(
            name='StorageCommitment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dir_type', models.CharField(choices=[('HOME', 'Home Directory'), ('PROJECT', 'Project Space'), ('SCRATCH', 'Scratch Space'), ('TEMP', 'Temporary Space')], max_length=16)),
                ('filesystem', models.CharField(max_length=64)),
                ('path', models.CharField(max_length=256)),
                ('commitment', models.BigIntegerField(blank=True, default=0, null=True)),
                ('registered', models.DateTimeField(auto_now_add=True)),
                ('created', models.DateTimeField(blank=True, null=True)),
                ('end_date', models.DateTimeField(blank=True, null=True)),
                ('reclaimed', models.DateTimeField(blank=True, null=True)),
                ('uid', models.IntegerField(blank=True, default=0)),
                ('gid', models.IntegerField(blank=True, default=0)),
                ('pid', models.IntegerField(blank=True, default=0)),
                ('permissions', models.CharField(default='0700', max_length=8)),
                ('is_purged', models.BooleanField(blank=True, default=True)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Project')),
                ('transactions', models.ManyToManyField(blank=True, to='openacct.Transaction')),
            ],
        ),        
        migrations.AddField(
            model_name='account',
            name='services',
            field=models.ManyToManyField(blank=True, to='openacct.Service'),
        ),
        migrations.AddField(
            model_name='service',
            name='system',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.System'),
        ),
        migrations.AddField(
            model_name='project',
            name='pi',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.User'),
        ),
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('queued', models.DateTimeField()),
                ('started', models.DateTimeField(blank=True, null=True)),
                ('completed', models.DateTimeField(blank=True, null=True)),
                ('jobid', models.CharField(max_length=32, unique=True)),
                ('name', models.CharField(blank=True, default='', max_length=64)),
                ('submit_host', models.CharField(blank=True, default='', max_length=64)),
                ('host_list', models.CharField(blank=True, default='', max_length=1024)),
                ('qos', models.CharField(blank=True, default='', max_length=64)),
                ('job_script', models.TextField(blank=True, default='')),
                ('wall_requested', models.IntegerField()),
                ('wall_duration', models.IntegerField(blank=True)),
                ('transactions', models.ManyToManyField(blank=True, to='openacct.Transaction')),
            ],
        ),
        migrations.AddField(
            model_name='account',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openacct.Project'),
        ),
    ]



"""
    openacct.shortcuts
    ~~~~~~~~~~~~~~~~~~

    This module provides a number of helper functions for performing 
    common operations on OpenAcct objects.
"""
from datetime import timedelta

from django.utils.timezone  import now

from .models    import (User, Project, UserProjectEvent, Account, System, Service, 
                        Transaction, Job, StorageCommitment)


def create_project(name, pi, description="", ldap_group="", do_create_account=True,
        account_name=None, account_duration=timedelta(days=365)):
    """Create and return a Project with the given traits. ``name`` should be the name 
    of project to create, and ``pi`` must either be an instance of User, or the string 
    name of a User in the database. 
    
    Optional ``description`` and ``ldap_group`` parameters set those respective 
    attributes on the created project.

    If ``do_create_account`` is True, an Account will be 
    created by calling ``create_account`` with ``account_name`` and ``account_duration`` 
    as its parameters, otherwise those parameters are ignored.
    """
    pi = pi if isinstance(pi, User) else User.objects.get(name=pi)
    project = Project.objects.create(name=name, pi=pi, description=description, ldap_group=ldap_group)
    add_user_to_project(pi, project)
    if do_create_account:
        create_account(project, name=account_name, duration=account_duration)
    return project


def create_account(project, name=None, duration=timedelta(days=365)):
    """Create and return an Account for the given Project. ``project`` must either 
    be an instance of Project, or the string name of a Project in the database. 

    If ``name`` is not provided, the Account will be named after the Project followed 
    by a dash and an index number. (For example, The first account for a project named 
    "test" will be named "test-1", the second will be "test-2", etc.)
    """
    project = project if isinstance(project, Project) else Project.objects.get(name=project)
    name = name if name else '{0}-{1}'.format(project.name, Account.next_index(project.name+'-'))
    return Account.objects.create(
        name=name, project=project, expires=now() + duration
    )


def add_user_to_project(user, project):
    """Add the given User to the given Project. Each argument can either be an instance of their
    respective objects, or the string name of a database object of their respective types.
    """
    user = user if isinstance(user, User) else User.objects.get(name=user)
    project = project if isinstance(project, Project) else Project.objects.get(name=project)
    UserProjectEvent.objects.create(user=user, project=project, event_type='ADDED')
    user.projects.add(project)


def remove_user_from_project(user, project):
    """Remove the given User from the given Project. Each argument can either be an instance of 
    their respective objects, or the string name of a database object of their respective types.
    """
    user = user if isinstance(user, User) else User.objects.get(name=user)
    project = project if isinstance(project, Project) else Project.objects.get(name=project)
    UserProjectEvent.objects.create(user=user, project=project, event_type='REMOVED')
    user.projects.remove(project)


def grant_service_access(service, account=None, project=None):
    """Grant access to a given Service to a single Account, or all active accounts on a Project.

    Must provide either the ``account`` or ``project`` argument. Each argument can either be an 
    instance of their respective objects, or the string name of a database object of their 
    respective types.
    """
    service = service if isinstance(service, Service) else Service.objects.get(name=service)
    if account:
        account = account if isinstance(account, Account) else Account.objects.get(name=account)
        account.services.add(service)
        Transaction.objects.create(
            service=service, account=account, tx_type='GRANT', 
            creator=account.project.pi, amt_used=0.0
        )
    elif project:
        project = project if isinstance(project, Project) else Project.objects.get(name=project)
        for acct in Account.objects.filter(project=project, active=True):
            acct.services.add(service)
            Transaction.objects.create(
                service=service, account=acct, tx_type='GRANT', 
                creator=project.pi, amt_used=0.0
            )
    else:
        raise TypeError('Must provide either project or account')


def revoke_service_access(service, account=None, project=None):
    """Revoke access to a given Service from a single Account, or all active accounts on a Project.

    Must provide either the ``account`` or ``project`` argument. Each argument can either be an 
    instance of their respective objects, or the string name of a database object of their 
    respective types.
    """
    service = service if isinstance(service, Service) else Service.objects.get(name=service)
    if account:
        account = account if isinstance(account, Account) else Account.objects.get(name=account)
        account.services.remove(service)
        Transaction.objects.create(
            service=service, account=account, tx_type='REVOKE', 
            creator=account.project.pi, amt_used=0.0
        )
    elif project:
        project = project if isinstance(project, Project) else Project.objects.get(name=project)
        for acct in Account.objects.filter(project=project, active=True):
            acct.services.remove(service)
            Transaction.objects.create(
                service=service, account=acct, tx_type='REVOKE', 
                creator=project.pi, amt_used=0.0
            )
    else:
        raise TypeError('Must provide either project or account')


def record_transaction(amt_used, service, user, account=None, project=None, 
        active=True, tx_type='DEBIT', amt_charged=0.0):
    """Create and return a transaction with the given parameters.

    Must provide either the ``account`` or ``project`` argument. ``service``, 
    ``user``, ``account`` and ``project`` can either be an instance of their 
    respective objects, or the string name of a database object of their 
    respective types.
    """
    service = service if isinstance(service, Service) else Service.objects.get(name=service)
    user = user if isinstance(user, User) else User.objects.get(name=user)
    if account:
        account = account if isinstance(account, Account) else Account.objects.get(name=account)
    elif project:
        account = Account.objects.filter(project=project, active=True).order_by('-created').first()
    else:
        raise TypeError('Must provide either project or account')
    
    return Transaction.objects.create(
        amt_used=amt_used, service=service, creator=user, account=account, active=active,
        tx_type=tx_type, amt_charged=amt_charged
    )


def record_job(jobid, queued, wall_requested, name='', submit_host='', host_list='', 
        qos='', job_script='', started=None, completed=None, wall_duration=None, 
        transactions=[]):
    """Create and return a job with the given parameters."""
    kwargs = {
        'jobid': jobid, 'queued': queued, 'wall_requested': wall_requested, 
        'name': name, 'submit_host': submit_host, 'host_list': host_list, 
        'qos': qos, 'job_script': job_script
    }
    if started: kwargs['started'] = started
    if completed: kwargs['completed'] = completed
    if wall_duration: kwargs['wall_duration'] = wall_duration

    job = Job.objects.create(**kwargs)
    if transactions:
        job.transactions.add(transactions)
    return job

